.model-parameter {
    margin-bottom: 15px;
}

div.tooltip {
    position: absolute;
    text-align: center;
    padding: 1px;
    font: 20px sans-serif;
    background: lightsteelblue;
    border: 3px;
    border-radius: 8px;
    pointer-events: none;
}

canvas.world-grid {
    position:absolute;
    border:1px dotted
}
